#!/usr/bin/python
# Copyright (C) 2010 W. Trevor King <wking@drexel.edu>
#
# This file is part of Dirtag.
#
# Dirtag is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Dirtag is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Dirtag.  If not, see <http://www.gnu.org/licenses/>.

import os
import os.path


class Node (list):
    def __init__(self, root, type, tags, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        self.root = root
        self.type = type
        self.tags = tags

    def pre_extend(self, a):
        """
        >>> n = Node(None, None, None, ['a', 'b', 'c'])
        >>> n.pre_extend([1, 2, 3])
        >>> print n
        [1, 2, 3, 'a', 'b', 'c']
        """
        for i,element in enumerate(a):
            self.insert(i, element)


class Tree (list):
    """
    >>> t = Tree('a')
    >>> t.append(Tree('1'))
    >>> t.append(Tree('2'))
    >>> t[0].append(Tree('A'))
    >>> print '\\n'.join(['|'.join(x) for x in t.traverse()])
    a
    a|1
    a|1|A
    a|2
    >>> print '\\n'.join(['|'.join(x) for x in t.traverse(prefix=['z'])])
    z|a
    z|a|1
    z|a|1|A
    z|a|2
    """
    def __init__(self, data=None, root=None, *args, **kwargs):
        self.data = data
        self.root = root
        list.__init__(self, *args, **kwargs)

    def traverse(self, prefix=[], depth=0, type='both', dirtag=None):
        p = prefix + [self.data]
        if depth < len(p) and type in ['both', 'dirs']:
            yield Node(self.root, 'dir', [], p[depth:])
        for child in self:
            if hasattr(child, 'traverse'):
                for grandchild in child.traverse(
                    p, depth=depth, type=type, dirtag=dirtag):
                    yield grandchild
            elif depth <= len(p) and type in ['both', 'files']:
                n = Node(self.root, 'file', [], (p + [child])[depth:])
                if dirtag != None:
                    n.tags = dirtag.node_tags(n)
                yield n


def dir_tree(root_dir):
    """Generate a directory tree.

    >>> t = dir_tree('test/tag')
    >>> print '\\n'.join(['|'.join(x) for x in t.traverse(depth=1)])
    1
    1|a1.dat
    1|b1.dat
    x
    x|y
    x|y|b1.dat
    x|a2.dat
    x|b3.dat
    >>> print '\\n'.join(['|'.join(x) for x in t.traverse(depth=1, type='files')])
    1|a1.dat
    1|b1.dat
    x|y|b1.dat
    x|a2.dat
    x|b3.dat
    >>> print '\\n'.join(['|'.join(x) for x in t.traverse(depth=1, type='dirs')])
    1
    x
    x|y
    """
    if not os.path.exists(root_dir):
        raise ValueError(root_dir)
    trees = {}
    for dirpath,dirnames,filenames in os.walk(root_dir, followlinks=True):
        t = trees.get(
            dirpath,
            Tree(os.path.basename(dirpath), root=root_dir))
        if dirpath == root_dir:
            root = t
        for d in sorted(dirnames):
            t2 = Tree(d, root=root_dir)
            trees[os.path.join(dirpath, d)] = t2
            t.append(t2)
        t.extend(sorted(filenames))
    return root


class Dirtag (object):
    """
    >>> d = Dirtag('test/raw', 'test/tag')
    >>> print '\\n'.join(['\t'.join([x.type, x.root, '|'.join(x),
    ...                              ','.join(['/'.join(t) for t in x.tags])])
    ...                   for x in d.elements()]
    ...                 )  # doctest: +NORMALIZE_WHITESPACE
    dir       test/raw      a      
    file      test/raw      a|a1.dat      1
    file      test/raw      a|a2.dat      x
    file      test/raw      a|a3.dat      
    dir       test/raw      b      
    file      test/raw      b|b1.dat      1,x/y
    file      test/raw      b|b2.dat      
    file      test/raw      b|b3.dat      x
    >>> print '\\n'.join(['|'.join(x) for x in d.elements(['x'])])
    x|y
    x|y|b1.dat
    x|a2.dat
    x|b3.dat
    >>> print '\\n'.join(['|'.join(x) for x in d.tags(['b', 'b2.dat'])])
    <BLANKLINE>
    >>> print '\\n'.join(['|'.join(x) for x in d.tags(['b', 'b1.dat'])])
    1
    x|y

    >>> print d.tag_path(['a', 'a3.dat'], ['x', 'y'])
    test/tag/x/y/a3.dat
    >>> os.listdir('test/tag/x/y')
    ['b1.dat']
    >>> d.add_tag(['a', 'a3.dat'], ['x', 'y'])
    >>> os.listdir('test/tag/x/y')
    ['a3.dat', 'b1.dat']
    >>> print 'Z'+os.path.realpath('test/tag/x/y/a3.dat') # doctest: +ELLIPSIS
    Z.../test/raw/a/a3.dat
    >>> d.remove_tag(['a', 'a3.dat'], ['x', 'y'])
    >>> os.listdir('test/tag/x/y')
    ['b1.dat']
    """
    def __init__(self, raw_dir, tag_dir):
        self.raw_dir = raw_dir
        self.tag_dir = tag_dir

    def elements(self, tag=None):
        if tag == None:
            tag = []
            nodes = dir_tree(self.raw_dir).traverse(depth=1, dirtag=self)
        else:
            p = [self.tag_dir]+tag
            nodes = dir_tree(os.path.join(*p)).traverse(
                prefix=p[:-1], depth=len(p), dirtag=self)
        for node in nodes:
            node.pre_extend(tag)
            yield node

    def tags(self, target):
        for node in dir_tree(self.tag_dir).traverse(depth=1):
            p = os.path.join(*([self.tag_dir]+node))
            t = os.path.abspath(os.path.join(*([self.raw_dir] + target)))
            if (os.path.islink(p) and os.path.realpath(p) == t):
                yield os.path.join(node[:-1])

    def tag_path(self, target, tag):
        # TODO: assumes unique basenames.  Check?
        return os.path.join(*([self.tag_dir] + tag + target[-1:]))

    def new_tag(self, tag):
        for i in range(len(tag)):
            os.mkdir(os.path.join(self.tag_dir, *tag[:i+1]))

    def add_tag(self, target, tag):
        tag_path = self.tag_path(target, tag)
        target_path = os.path.abspath(os.path.join(*([self.raw_dir]+target)))
        cwd = os.getcwd()
        try:
            os.chdir(os.path.dirname(tag_path))
            os.symlink(os.path.relpath(target_path),
                       os.path.basename(tag_path))
        finally:
            os.chdir(cwd)

    def remove_tag(self, target, tag):
        tag_path = self.tag_path(target, tag)
        cwd = os.getcwd()
        try:
            os.chdir(os.path.dirname(tag_path))
            os.remove(os.path.basename(tag_path))
        finally:
            os.chdir(cwd)

    def raw_node(self, node):
        """Return a node for the real file linked to by node.
        """
        if node.root == self.raw_dir:
            return node
        path = os.path.realpath(os.path.join(node.root, *node))
        raw_path = os.path.relpath(path, self.raw_dir)
        return raw_path.split(os.path.sep)

    def node_tags(self, node):
        """Return a list of tags for a particular node.
        """
        return self.tags(self.raw_node(node))


if __name__ == '__main__':
    import doctest
    doctest.testmod()
